from os import PathLike
from typing import Any, TypedDict


class ConfiguredTargetTuple(TypedDict):
    target: str
    toolchain_root: PathLike[Any]
    toolchain_sysroot: PathLike[Any] | None
    toolchain_flags: str
    binutils_flavor: str
    cc_flavor: str
    gcc_install_dir: PathLike[Any] | None



import importlib.metadata
from typing import Final, TYPE_CHECKING

import packaging.version

if TYPE_CHECKING:
    # pyright only works with semver 3.x
    from semver.version import Version
else:
    try:
        from semver.version import Version  # type: ignore[import-untyped,unused-ignore]
    except ModuleNotFoundError:
        # semver 2.x
        from semver import VersionInfo as Version  # type: ignore[import-untyped,unused-ignore]

# NOTE: one cannot print logs in the version helpers, because the version info
# is initialized so early (before argparse can look at argv because --version
# requires version info to be ready) that the porcelain status is not yet
# available.

_PYPI_PRERELEASE_KINDS_MAP: Final = {
    "a": "alpha",
    "b": "beta",
    "rc": "rc",
}


# based on https://python-semver.readthedocs.io/en/3.0.2/advanced/convert-pypi-to-semver.html
def _convert2semver(ver: packaging.version.Version) -> Version:
    if ver.epoch:
        raise ValueError("Can't convert an epoch to semver")
    if ver.post:
        raise ValueError("Can't convert a post part to semver")

    pre: str | None = None
    if ver.pre:
        kind, val = ver.pre
        pre = f"{_PYPI_PRERELEASE_KINDS_MAP.get(kind, kind)}.{val}"

    maj, min, pat = ver.release[:3]
    return Version(maj, min, pat, prerelease=pre, build=ver.dev)


def _init_pkg_semver() -> Version:
    pkg_pypi_ver = packaging.version.Version(importlib.metadata.version("ruyi"))
    # log.D(f"PyPI-style version of ruyi: {pkg_pypi_ver}")
    return _convert2semver(pkg_pypi_ver)


RUYI_SEMVER: Final = _init_pkg_semver()
RUYI_USER_AGENT: Final = f"ruyi/{RUYI_SEMVER}"

COPYRIGHT_NOTICE: Final = """\
Copyright (C) Institute of Software, Chinese Academy of Sciences (ISCAS).
All rights reserved.
License: Apache-2.0 <https://www.apache.org/licenses/LICENSE-2.0>
\
"""

MPL_REDIST_NOTICE: Final = """\
This distribution of ruyi contains code licensed under the Mozilla Public
License 2.0 (https://mozilla.org/MPL/2.0/). You can get the respective
project's sources from the project's official website:

* certifi: https://github.com/certifi/python-certifi
\
"""



from os import PathLike
from typing import Any, Iterable, TypeGuard, cast

from ..pluginhost.ctx import PluginHostContext, SupportsEvalFunction
from .pkg_manifest import EmulatorFlavor


class InvalidProfilePluginError(RuntimeError):
    def __init__(self, s: str) -> None:
        super().__init__(f"invalid arch profile plugin: {s}")


def validate_list_str(x: object) -> TypeGuard[list[str]]:
    if not isinstance(x, list):
        return False
    x = cast(list[object], x)
    return all(isinstance(y, str) for y in x)


def validate_list_str_or_none(x: object) -> TypeGuard[list[str] | None]:
    return True if x is None else validate_list_str(x)


def validate_dict_str_str(x: object) -> TypeGuard[dict[str, str]]:
    if not isinstance(x, dict):
        return False
    for k, v in cast(dict[object, object], x).items():
        if not isinstance(k, str) or not isinstance(v, str):
            return False
    return True


class PluginProfileProvider:
    def __init__(
        self,
        phctx: PluginHostContext[Any, SupportsEvalFunction],
        plugin_id: str,
    ) -> None:
        self._phctx = phctx
        self._plugin_id = plugin_id
        self._ev = phctx.make_evaluator()

    def _must_get(self, name: str) -> object:
        if v := self._phctx.get_from_plugin(self._plugin_id, name):
            return v
        raise InvalidProfilePluginError(
            f"'{name}' not found in plugin '{self._plugin_id}'"
        )

    def list_all_profile_ids(self) -> list[str]:
        fn = self._must_get("list_all_profile_ids_v1")
        ret = self._ev.eval_function(fn)
        if not validate_list_str(ret):
            raise InvalidProfilePluginError(
                "list_all_profile_ids must return list[str]"
            )

        return ret

    def list_needed_quirks(self, profile_id: str) -> list[str] | None:
        # For backward compatibility, try "list_needed_quirks_v1" first, then
        # fall back to "list_needed_flavors_v1" if the former is not available.
        fn = self._phctx.get_from_plugin(self._plugin_id, "list_needed_quirks_v1")
        if fn is None:
            fn = self._must_get("list_needed_flavors_v1")

        ret = self._ev.eval_function(fn, profile_id)
        if not validate_list_str_or_none(ret):
            raise InvalidProfilePluginError(
                "list_needed_quirks_v1 must return list[str] | None"
            )

        return ret

    def get_common_flags(self, profile_id: str, toolchain_quirks: list[str]) -> str:
        result = self._maybe_get_common_flags_v2(profile_id, toolchain_quirks)
        if result is not None:
            return result
        return self._get_common_flags_v1(profile_id)

    def _get_common_flags_v1(self, profile_id: str) -> str:
        fn = self._must_get("get_common_flags_v1")
        ret = self._ev.eval_function(fn, profile_id)
        if not isinstance(ret, str):
            raise InvalidProfilePluginError("get_common_flags_v1 must return str")

        return ret

    def _maybe_get_common_flags_v2(
        self,
        profile_id: str,
        toolchain_flavors: list[str],
    ) -> str | None:
        fn = self._phctx.get_from_plugin(self._plugin_id, "get_common_flags_v2")
        if fn is None:
            return None

        ret = self._ev.eval_function(fn, profile_id, toolchain_flavors)
        if not isinstance(ret, str):
            raise InvalidProfilePluginError("get_common_flags_v2 must return str")

        return ret

    def get_needed_emulator_pkg_flavors(
        self,
        profile_id: str,
        flavor: EmulatorFlavor,
    ) -> Iterable[str]:
        fn = self._must_get("get_needed_emulator_pkg_flavors_v1")
        ret = self._ev.eval_function(
            fn,
            profile_id,
            flavor,
        )
        if not validate_list_str(ret):
            raise InvalidProfilePluginError(
                "get_needed_emulator_pkg_flavors_v1 must return list[str]"
            )

        return ret

    def check_emulator_flavor(
        self,
        profile_id: str,
        flavor: EmulatorFlavor,
        emulator_pkg_flavors: list[str] | None,
    ) -> bool:
        fn = self._must_get("check_emulator_flavor_v1")
        ret = self._ev.eval_function(
            fn,
            profile_id,
            flavor,
            emulator_pkg_flavors,
        )
        if not isinstance(ret, bool):
            raise InvalidProfilePluginError("check_emulator_flavor_v1 must return bool")

        return ret

    def get_env_config_for_emu_flavor(
        self,
        profile_id: str,
        flavor: EmulatorFlavor,
        sysroot: PathLike[Any] | None,
    ) -> dict[str, str] | None:
        fn = self._must_get("get_env_config_for_emu_flavor_v1")
        ret = self._ev.eval_function(
            fn,
            profile_id,
            flavor,
            str(sysroot) if sysroot is not None else None,
        )
        if not validate_dict_str_str(ret):
            raise InvalidProfilePluginError(
                "get_env_config_for_emu_flavor_v1 must return dict[str, str]"
            )

        return ret


class ProfileProxy:
    def __init__(
        self,
        provider: PluginProfileProvider,
        arch: str,
        profile_id: str,
    ) -> None:
        self._provider = provider
        self._arch = arch
        self._id = profile_id

    @property
    def arch(self) -> str:
        return self._arch

    @property
    def id(self) -> str:
        return self._id

    @property
    def need_quirks(self) -> set[str]:
        r = self._provider.list_needed_quirks(self._id)
        return set(r) if r else set()

    def get_common_flags(self, toolchain_flavors: list[str]) -> str:
        return self._provider.get_common_flags(self._id, toolchain_flavors)

    def get_needed_emulator_pkg_flavors(
        self,
        flavor: EmulatorFlavor,
    ) -> set[str]:
        return set(self._provider.get_needed_emulator_pkg_flavors(self._id, flavor))

    def check_emulator_flavor(
        self,
        flavor: EmulatorFlavor,
        emulator_pkg_flavors: list[str] | None,
    ) -> bool:
        return self._provider.check_emulator_flavor(
            self._id, flavor, emulator_pkg_flavors
        )

    def get_env_config_for_emu_flavor(
        self,
        flavor: EmulatorFlavor,
        sysroot: PathLike[Any] | None,
    ) -> dict[str, str] | None:
        return self._provider.get_env_config_for_emu_flavor(self._id, flavor, sysroot)



import datetime
from functools import cached_property
import locale
import os.path
from os import PathLike
import pathlib
import sys
from typing import Any, Final, Iterable, Sequence, TypedDict, TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import NotRequired, Self

    from ..log import RuyiLogger
    from ..ruyipkg.repo import MetadataRepo
    from ..ruyipkg.state import RuyipkgGlobalStateStore
    from ..telemetry.provider import TelemetryProvider
    from ..utils.global_mode import ProvidesGlobalMode
    from .news import NewsReadStatusStore

from . import schema


if sys.platform == "linux":
    PRESET_GLOBAL_CONFIG_LOCATIONS: Final[list[str]] = [
        # TODO: enable distro packagers to customize the $PREFIX to suit their
        # particular FS layout if necessary.
        "/usr/share/ruyi/config.toml",
        "/usr/local/share/ruyi/config.toml",
    ]
else:
    PRESET_GLOBAL_CONFIG_LOCATIONS: Final[list[str]] = []

DEFAULT_APP_NAME: Final = "ruyi"
DEFAULT_REPO_URL: Final = "https://github.com/ruyisdk/packages-index.git"
DEFAULT_REPO_BRANCH: Final = "main"


def get_host_path_fragment_for_binary_install_dir(canonicalized_host: str) -> str:
    # e.g. linux/amd64 -> amd64; "windows/amd64" -> "windows-amd64"
    if canonicalized_host.startswith("linux/"):
        return canonicalized_host[6:]
    return canonicalized_host.replace("/", "-")


def _get_lang_code() -> str:
    lang = locale.getlocale()[0]
    return lang or "en_US"


class GlobalConfigPackagesType(TypedDict):
    prereleases: "NotRequired[bool]"


class GlobalConfigRepoType(TypedDict):
    local: "NotRequired[str]"
    remote: "NotRequired[str]"
    branch: "NotRequired[str]"


class GlobalConfigInstallationType(TypedDict):
    # Undocumented: whether this Ruyi installation is externally managed.
    #
    # Can be used by distro packagers (by placing a config file in /etc/xdg/ruyi)
    # to signify this status to an official Ruyi build (where IS_PACKAGED is
    # True), to prevent e.g. accidental self-uninstallation.
    externally_managed: "NotRequired[bool]"


class GlobalConfigTelemetryType(TypedDict):
    mode: "NotRequired[str]"
    upload_consent: "NotRequired[datetime.datetime | str]"
    pm_telemetry_url: "NotRequired[str]"


class GlobalConfigRootType(TypedDict):
    installation: "NotRequired[GlobalConfigInstallationType]"
    packages: "NotRequired[GlobalConfigPackagesType]"
    repo: "NotRequired[GlobalConfigRepoType]"
    telemetry: "NotRequired[GlobalConfigTelemetryType]"


class GlobalConfig:
    def __init__(self, gm: "ProvidesGlobalMode", logger: "RuyiLogger") -> None:
        from ..utils.xdg_basedir import XDGBaseDir

        self._gm = gm
        self.logger = logger

        # all defaults
        self.override_repo_dir: str | None = None
        self.override_repo_url: str | None = None
        self.override_repo_branch: str | None = None
        self.include_prereleases = False
        self.is_installation_externally_managed = False

        self._lang_code = _get_lang_code()

        self._dirs = XDGBaseDir(DEFAULT_APP_NAME)

        self._telemetry_mode: str | None = None
        self._telemetry_upload_consent: datetime.datetime | None = None
        self._telemetry_pm_telemetry_url: str | None = None

    def apply_config(self, config_data: GlobalConfigRootType) -> None:
        if ins_cfg := config_data.get(schema.SECTION_INSTALLATION):
            self.is_installation_externally_managed = ins_cfg.get(
                schema.KEY_INSTALLATION_EXTERNALLY_MANAGED,
                False,
            )

        if pkgs_cfg := config_data.get(schema.SECTION_PACKAGES):
            self.include_prereleases = pkgs_cfg.get(
                schema.KEY_PACKAGES_PRERELEASES, False
            )

        if repo_cfg := config_data.get(schema.SECTION_REPO):
            self.override_repo_dir = repo_cfg.get(schema.KEY_REPO_LOCAL, None)
            self.override_repo_url = repo_cfg.get(schema.KEY_REPO_REMOTE, None)
            self.override_repo_branch = repo_cfg.get(schema.KEY_REPO_BRANCH, None)

            if self.override_repo_dir:
                if not pathlib.Path(self.override_repo_dir).is_absolute():
                    self.logger.W(
                        f"the local repo path '{self.override_repo_dir}' is not absolute; ignoring"
                    )
                    self.override_repo_dir = None

        if tele_cfg := config_data.get(schema.SECTION_TELEMETRY):
            self._telemetry_mode = tele_cfg.get(schema.KEY_TELEMETRY_MODE, None)
            self._telemetry_pm_telemetry_url = tele_cfg.get(
                schema.KEY_TELEMETRY_PM_TELEMETRY_URL,
                None,
            )

            self._telemetry_upload_consent = None
            if consent := tele_cfg.get(schema.KEY_TELEMETRY_UPLOAD_CONSENT, None):
                if isinstance(consent, datetime.datetime):
                    self._telemetry_upload_consent = consent

    def get_by_key(self, key: str | Sequence[str]) -> object | None:
        parsed_key = schema.parse_config_key(key)
        section, sel = parsed_key[0], parsed_key[1:]
        if section == schema.SECTION_INSTALLATION:
            return self._get_section_installation(sel)
        elif section == schema.SECTION_PACKAGES:
            return self._get_section_packages(sel)
        elif section == schema.SECTION_REPO:
            return self._get_section_repo(sel)
        elif section == schema.SECTION_TELEMETRY:
            return self._get_section_telemetry(sel)
        else:
            return None

    def _get_section_installation(self, selector: list[str]) -> object | None:
        if len(selector) != 1:
            return None
        leaf = selector[0]
        if leaf == schema.KEY_INSTALLATION_EXTERNALLY_MANAGED:
            return self.is_installation_externally_managed
        else:
            return None

    def _get_section_packages(self, selector: list[str]) -> object | None:
        if len(selector) != 1:
            return None
        leaf = selector[0]
        if leaf == schema.KEY_PACKAGES_PRERELEASES:
            return self.include_prereleases
        else:
            return None

    def _get_section_repo(self, selector: list[str]) -> object | None:
        if len(selector) != 1:
            return None
        leaf = selector[0]
        if leaf == schema.KEY_REPO_BRANCH:
            return self.override_repo_branch
        elif leaf == schema.KEY_REPO_LOCAL:
            return self.override_repo_dir
        elif leaf == schema.KEY_REPO_REMOTE:
            return self.override_repo_url
        else:
            return None

    def _get_section_telemetry(self, selector: list[str]) -> object | None:
        if len(selector) != 1:
            return None
        leaf = selector[0]
        if leaf == schema.KEY_TELEMETRY_MODE:
            return self.telemetry_mode
        elif leaf == schema.KEY_TELEMETRY_PM_TELEMETRY_URL:
            return self.override_pm_telemetry_url
        elif leaf == schema.KEY_TELEMETRY_UPLOAD_CONSENT:
            return self.telemetry_upload_consent_time
        else:
            return None

    @property
    def argv0(self) -> str:
        return self._gm.argv0

    @property
    def main_file(self) -> str:
        return self._gm.main_file

    @property
    def self_exe(self) -> str:
        return self._gm.self_exe

    @property
    def is_debug(self) -> bool:
        return self._gm.is_debug

    @property
    def is_experimental(self) -> bool:
        return self._gm.is_experimental

    @property
    def is_packaged(self) -> bool:
        return self._gm.is_packaged

    @property
    def is_porcelain(self) -> bool:
        return self._gm.is_porcelain

    @property
    def is_telemetry_optout(self) -> bool:
        return self._gm.is_telemetry_optout

    @property
    def is_cli_autocomplete(self) -> bool:
        return self._gm.is_cli_autocomplete

    @property
    def venv_root(self) -> str | None:
        return self._gm.venv_root

    @property
    def lang_code(self) -> str:
        return self._lang_code

    @property
    def cache_root(self) -> os.PathLike[Any]:
        return self._dirs.app_cache

    @property
    def data_root(self) -> os.PathLike[Any]:
        return self._dirs.app_data

    @property
    def state_root(self) -> os.PathLike[Any]:
        return self._dirs.app_state

    @cached_property
    def news_read_status(self) -> "NewsReadStatusStore":
        from .news import NewsReadStatusStore

        filename = os.path.join(self.ensure_state_dir(), "news.read.txt")
        return NewsReadStatusStore(filename)

    @property
    def telemetry_root(self) -> os.PathLike[Any]:
        return pathlib.Path(self.ensure_state_dir()) / "telemetry"

    @cached_property
    def telemetry(self) -> "TelemetryProvider | None":
        from ..telemetry.provider import TelemetryProvider

        return None if self.telemetry_mode == "off" else TelemetryProvider(self)

    @property
    def telemetry_mode(self) -> str:
        return self._telemetry_mode or "on"

    @property
    def telemetry_upload_consent_time(self) -> datetime.datetime | None:
        return self._telemetry_upload_consent

    @property
    def override_pm_telemetry_url(self) -> str | None:
        return self._telemetry_pm_telemetry_url

    def get_repo_dir(self) -> str:
        return self.override_repo_dir or os.path.join(self.cache_root, "packages-index")

    def get_repo_url(self) -> str:
        return self.override_repo_url or DEFAULT_REPO_URL

    def get_repo_branch(self) -> str:
        return self.override_repo_branch or DEFAULT_REPO_BRANCH

    @cached_property
    def repo(self) -> "MetadataRepo":
        from ..ruyipkg.repo import MetadataRepo

        return MetadataRepo(self)

    def ensure_distfiles_dir(self) -> str:
        path = pathlib.Path(self.ensure_cache_dir()) / "distfiles"
        path.mkdir(parents=True, exist_ok=True)
        return str(path)

    def global_binary_install_root(self, host: str, slug: str) -> str:
        host_path = get_host_path_fragment_for_binary_install_dir(host)
        path = pathlib.Path(self.ensure_data_dir()) / "binaries" / host_path / slug
        return str(path)

    def global_blob_install_root(self, slug: str) -> str:
        path = pathlib.Path(self.ensure_data_dir()) / "blobs" / slug
        return str(path)

    def lookup_binary_install_dir(self, host: str, slug: str) -> PathLike[Any] | None:
        host_path = get_host_path_fragment_for_binary_install_dir(host)
        for data_dir in self._dirs.app_data_dirs:
            p = data_dir / "binaries" / host_path / slug
            if p.exists():
                return p
        return None

    @property
    def ruyipkg_state_root(self) -> os.PathLike[Any]:
        return pathlib.Path(self.ensure_state_dir()) / "ruyipkg"

    @cached_property
    def ruyipkg_global_state(self) -> "RuyipkgGlobalStateStore":
        from ..ruyipkg.state import RuyipkgGlobalStateStore

        return RuyipkgGlobalStateStore(self.ruyipkg_state_root)

    def ensure_data_dir(self) -> os.PathLike[Any]:
        p = self._dirs.app_data
        p.mkdir(parents=True, exist_ok=True)
        return p

    def ensure_cache_dir(self) -> os.PathLike[Any]:
        p = self._dirs.app_cache
        p.mkdir(parents=True, exist_ok=True)
        return p

    def ensure_config_dir(self) -> os.PathLike[Any]:
        p = self._dirs.app_config
        p.mkdir(parents=True, exist_ok=True)
        return p

    def ensure_state_dir(self) -> os.PathLike[Any]:
        p = self._dirs.app_state
        p.mkdir(parents=True, exist_ok=True)
        return p

    def iter_preset_configs(self) -> Iterable[os.PathLike[Any]]:
        """
        Yields possible Ruyi config files in all preset config path locations,
        sorted by precedence from lowest to highest (so that each file may be
        simply applied consecutively).
        """

        for path in PRESET_GLOBAL_CONFIG_LOCATIONS:
            yield pathlib.Path(path)

    def iter_xdg_configs(self) -> Iterable[os.PathLike[Any]]:
        """
        Yields possible Ruyi config files in all XDG config paths, sorted by precedence
        from lowest to highest (so that each file may be simply applied consecutively).
        """

        for config_dir in reversed(list(self._dirs.app_config_dirs)):
            yield config_dir / "config.toml"

    @property
    def local_user_config_file(self) -> pathlib.Path:
        return self._dirs.app_config / "config.toml"

    def try_apply_config_file(self, path: os.PathLike[Any]) -> None:
        import tomlkit

        try:
            with open(path, "rb") as fp:
                data: Any = tomlkit.load(fp)
        except FileNotFoundError:
            return

        self.logger.D(f"applying config: {data}")
        self.apply_config(data)

    @classmethod
    def load_from_config(cls, gm: "ProvidesGlobalMode", logger: "RuyiLogger") -> "Self":
        obj = cls(gm, logger)

        for config_path in obj.iter_preset_configs():
            obj.logger.D(f"trying config file from preset location: {config_path}")
            obj.try_apply_config_file(config_path)

        for config_path in obj.iter_xdg_configs():
            obj.logger.D(f"trying config file from XDG path: {config_path}")
            obj.try_apply_config_file(config_path)

        # let environment variable take precedence
        if gm.is_telemetry_optout:
            obj._telemetry_mode = "off"

        return obj



import calendar
import datetime
import functools
import json
import pathlib
import sys
import time
from typing import Callable, TYPE_CHECKING, cast
import uuid

from ..log import RuyiLogger
from .node_info import NodeInfo, gather_node_info
from .scope import TelemetryScope
from .store import TelemetryStore

if TYPE_CHECKING:
    # for avoiding circular import
    from ..config import GlobalConfig

FALLBACK_PM_TELEMETRY_ENDPOINT = "https://api.ruyisdk.cn/telemetry/pm/"

FIRST_RUN_PROMPT = """\
Welcome to RuyiSDK! This appears to be your first run of [yellow]ruyi[/].

By default, the RuyiSDK team collects anonymous usage data to help us improve
the product. No personal information or detail about your project is ever
collected. The data will be uploaded to RuyiSDK team-managed servers located
in the Chinese mainland if you agree to the uploading. You can change this
setting at any time by running [yellow]ruyi telemetry consent[/] or
[yellow]ruyi telemetry optout[/].

We would like to ask if you agree to have basic information about this [yellow]ruyi[/]
installation uploaded, right now, for one time. This will allow the RuyiSDK team
to have more precise knowledge about the product's adoption. Thank you for
your support!
"""


def next_utc_weekday(wday: int, now: float | None = None) -> int:
    t = time.gmtime(now)
    mday_delta = wday - t.tm_wday
    if mday_delta < 0:
        mday_delta += 7

    next_t = (
        t.tm_year,
        t.tm_mon,
        t.tm_mday + mday_delta,
        0,  # tm_hour
        0,  # tm_min
        0,  # tm_sec
        0,  # tm_wday
        0,  # tm_yday
        -1,  # tm_isdst
    )
    return calendar.timegm(next_t)


def set_telemetry_mode(
    gc: "GlobalConfig",
    mode: str,
    consent_time: datetime.datetime | None = None,
    show_cli_feedback: bool = True,
) -> None:
    """Set telemetry mode and consent time (if applicable) in the user preference."""

    from ..config.editor import ConfigEditor
    from ..config import schema

    logger = gc.logger

    with ConfigEditor.work_on_user_local_config(gc) as ed:
        ed.set_value((schema.SECTION_TELEMETRY, schema.KEY_TELEMETRY_MODE), mode)

        if mode == "on":
            if consent_time is None:
                consent_time = datetime.datetime.now().astimezone()
            ed.set_value(
                (schema.SECTION_TELEMETRY, schema.KEY_TELEMETRY_UPLOAD_CONSENT),
                consent_time,
            )
        else:
            ed.unset_value(
                (schema.SECTION_TELEMETRY, schema.KEY_TELEMETRY_UPLOAD_CONSENT)
            )

        ed.stage()

    if not show_cli_feedback:
        return
    match mode:
        case "on":
            logger.I("telemetry data uploading is now enabled")
            logger.I(
                "you can opt out at any time by running [yellow]ruyi telemetry optout[/]"
            )
        case "local":
            logger.I("telemetry mode is now set to local collection only")
            logger.I(
                "you can re-enable telemetry data uploading at any time by running [yellow]ruyi telemetry consent[/]"
            )
            logger.I(
                "or opt out at any time by running [yellow]ruyi telemetry optout[/]"
            )
        case "off":
            logger.I("telemetry data collection is now disabled")
            logger.I(
                "you can re-enable telemetry data uploads at any time by running [yellow]ruyi telemetry consent[/]"
            )
        case _:
            raise ValueError(f"invalid telemetry mode: {mode}")


class TelemetryProvider:
    def __init__(self, gc: "GlobalConfig") -> None:
        self.state_root = pathlib.Path(gc.telemetry_root)
        self.local_mode = gc.telemetry_mode == "local"
        self.upload_consent_time = gc.telemetry_upload_consent_time

        self._discard_events = False
        self._gc = gc
        self._is_first_run = False
        self._stores: dict[TelemetryScope, TelemetryStore] = {}
        self._upload_on_exit = False

        # create the PM store
        self.init_store(TelemetryScope(None))
        # TODO: add real multi-repo support
        self.init_store(TelemetryScope("ruyisdk"))

    @property
    def logger(self) -> RuyiLogger:
        return self._gc.logger

    def store(self, scope: TelemetryScope) -> TelemetryStore | None:
        return self._stores.get(scope)

    def init_store(self, scope: TelemetryScope) -> None:
        store_root = self.state_root
        api_url_fn: Callable[[], str | None] | None = None
        if repo_name := scope.repo_name:
            if repo_name != "ruyisdk":
                raise NotImplementedError("multi-repo support not implemented yet")
            store_root = store_root / "repos" / repo_name

            def _f() -> str | None:
                # access the repo attribute lazily to speed up CLI startup
                return self._gc.repo.get_telemetry_api_url("repo")

            api_url_fn = _f
        else:
            # configure the PM telemetry endpoint
            api_url_fn = functools.partial(self._detect_pm_api_url, self._gc)

        store = TelemetryStore(
            self.logger,
            scope,
            store_root,
            api_url_factory=api_url_fn,
        )
        self._stores[scope] = store

    def _detect_pm_api_url(self, gc: "GlobalConfig") -> str | None:
        url = FALLBACK_PM_TELEMETRY_ENDPOINT
        cfg_src = "fallback"
        if gc.override_pm_telemetry_url is not None:
            cfg_src = "local config"
            url = gc.override_pm_telemetry_url
        else:
            if repo_provided_url := gc.repo.get_telemetry_api_url("pm"):
                cfg_src = "repo"
                url = repo_provided_url
        self.logger.D(
            f"configured PM telemetry endpoint via {cfg_src}: {url or '(n/a)'}"
        )
        return url

    @property
    def installation_file(self) -> pathlib.Path:
        return self.state_root / "installation.json"

    def check_first_run_status(self) -> None:
        """Check if this is the first run of the application by checking if installation file exists.
        This must be done before init_installation() is potentially called.
        """
        self._is_first_run = not self.installation_file.exists()

    @property
    def is_first_run(self) -> bool:
        """Check if this is the first run of the application."""
        return self._is_first_run

    def init_installation(self, force_reinit: bool) -> NodeInfo | None:
        installation_file = self.installation_file
        if installation_file.exists() and not force_reinit:
            return self.read_installation_data()

        # either this is a fresh installation or we're forcing a refresh
        installation_id = uuid.uuid4()
        self.logger.D(
            f"initializing telemetry data store, installation_id={installation_id.hex}"
        )
        self.state_root.mkdir(parents=True, exist_ok=True)

        # (over)write installation data
        installation_data = gather_node_info(installation_id)
        with open(installation_file, "wb") as fp:
            fp.write(json.dumps(installation_data).encode("utf-8"))
        return installation_data

    def read_installation_data(self) -> NodeInfo | None:
        with open(self.installation_file, "rb") as fp:
            return cast(NodeInfo, json.load(fp))

    def upload_weekday(self) -> int | None:
        try:
            installation_data = self.read_installation_data()
        except FileNotFoundError:
            # init the node info if it's gone
            installation_data = self.init_installation(False)

        if installation_data is None:
            return None

        try:
            report_uuid_prefix = int(installation_data["report_uuid"][:8], 16)
        except ValueError:
            return None

        return report_uuid_prefix % 7  # 0 is Monday

    def has_upload_consent(self, time_now: float | None = None) -> bool:
        if self.upload_consent_time is None:
            return False
        if time_now is None:
            time_now = time.time()
        return self.upload_consent_time.timestamp() <= time_now

    def print_telemetry_notice(self, for_cli_verbose_output: bool = False) -> None:
        if self.local_mode:
            if for_cli_verbose_output:
                self.logger.I(
                    "telemetry mode is [green]local[/]: local data collection only, no uploads"
                )
            return

        now = time.time()
        if self.has_upload_consent(now) and not for_cli_verbose_output:
            self.logger.D("user has consented to telemetry upload")
            return

        upload_wday = self.upload_weekday()
        if upload_wday is None:
            return
        upload_wday_name = calendar.day_name[upload_wday]

        next_upload_day_ts = next_utc_weekday(upload_wday, now)
        next_upload_day = time.localtime(next_upload_day_ts)
        next_upload_day_end = time.localtime(next_upload_day_ts + 86400)
        next_upload_day_str = time.strftime("%Y-%m-%d %H:%M:%S %z", next_upload_day)
        next_upload_day_end_str = time.strftime(
            "%Y-%m-%d %H:%M:%S %z", next_upload_day_end
        )

        today_is_upload_day = self.is_upload_day(now)
        if for_cli_verbose_output:
            self.logger.I(
                "telemetry mode is [green]on[/]: data is collected and periodically uploaded"
            )
            self.logger.I(
                f"non-tracking usage information will be uploaded to RuyiSDK-managed servers [bold green]every {upload_wday_name}[/]"
            )
        else:
            self.logger.W(
                f"this [yellow]ruyi[/] installation has telemetry mode set to [yellow]on[/], and [bold]will upload non-tracking usage information to RuyiSDK-managed servers[/] [bold green]every {upload_wday_name}[/]"
            )

        if today_is_upload_day:
            for scope, store in self._stores.items():
                has_uploaded_today = self.has_uploaded_today(scope, now)
                if has_uploaded_today:
                    if last_upload_time := store.last_upload_timestamp:
                        last_upload_time_str = time.strftime(
                            "%Y-%m-%d %H:%M:%S %z", time.localtime(last_upload_time)
                        )
                        self.logger.I(
                            f"scope {scope}: usage information has already been uploaded today at {last_upload_time_str}"
                        )
                    else:
                        self.logger.I(
                            f"scope {scope}: usage information has already been uploaded sometime today"
                        )
                else:
                    self.logger.I(
                        f"scope {scope}: the next upload will happen [bold green]today[/] if not already"
                    )
        else:
            self.logger.I(
                f"the next upload will happen anytime [yellow]ruyi[/] is executed between [bold green]{next_upload_day_str}[/] and [bold green]{next_upload_day_end_str}[/]"
            )

        if not for_cli_verbose_output:
            self.logger.I("in order to hide this banner:")
            self.logger.I("- opt out with [yellow]ruyi telemetry optout[/]")
            self.logger.I("- or give consent with [yellow]ruyi telemetry consent[/]")

    def next_upload_day(self, time_now: float | None = None) -> int | None:
        upload_wday = self.upload_weekday()
        if upload_wday is None:
            return None
        return next_utc_weekday(upload_wday, time_now)

    def is_upload_day(self, time_now: float | None = None) -> bool:
        if time_now is None:
            time_now = time.time()
        if upload_day := self.next_upload_day(time_now):
            return upload_day <= time_now
        return False

    def has_uploaded_today(
        self,
        scope: TelemetryScope,
        time_now: float | None = None,
    ) -> bool:
        if time_now is None:
            time_now = time.time()
        if upload_day := self.next_upload_day(time_now):
            upload_day_end = upload_day + 86400
            store = self.store(scope)
            if store is None:
                return False
            if last_upload_time := store.last_upload_timestamp:
                return upload_day <= last_upload_time < upload_day_end
        return False

    def record(self, scope: TelemetryScope, kind: str, **params: object) -> None:
        if store := self.store(scope):
            return store.record(kind, **params)
        self.logger.D(f"no telemetry store for scope {scope}, discarding event")

    def discard_events(self, v: bool = True) -> None:
        self._discard_events = v

    def flush(self, *, upload_now: bool = False) -> None:
        now = time.time()

        # We may be self-uninstalling and purging all state data, and in this
        # case we don't want to record anything (thus re-creating directories).
        if self._discard_events:
            self.logger.D("discarding collected telemetry data")
            return

        self.logger.D("flushing telemetry to persistent store")

        for scope, store in self._stores.items():
            store.persist(now)

            # try to upload if forced (upload_now or _upload_on_exit), or:
            #
            # * we're not in local mode
            # * today is the day
            # * we haven't uploaded today
            if not (upload_now or self._upload_on_exit) and (
                self.local_mode
                or not self.is_upload_day(now)
                or self.has_uploaded_today(scope, now)
            ):
                continue

            self.prepare_data_for_upload(store)
            self.upload_staged_payloads(store)

    def prepare_data_for_upload(self, store: TelemetryStore) -> None:
        installation_data: NodeInfo | None = None
        if store.scope.is_pm:
            try:
                installation_data = self.read_installation_data()
            except FileNotFoundError:
                # should not happen due to is_upload_day() initializing it for us
                # beforehand, but proceed without node info nonetheless
                pass

        return store.prepare_data_for_upload(installation_data)

    def upload_staged_payloads(self, store: TelemetryStore) -> None:
        if self.local_mode:
            return

        return store.upload_staged_payloads()

    def maybe_prompt_for_first_run_upload(self) -> None:
        """
        Ask whether the user consents to a first-run telemetry upload when
        running for the first time (OOBE) and with an interactive stdin.
        """

        # Only prompt if this is first run and stdin is a TTY
        if not (self.is_first_run and sys.stdin.isatty()):
            return

        from ..cli import user_input

        self.logger.I(FIRST_RUN_PROMPT)
        if not user_input.ask_for_yesno_confirmation(
            self.logger,
            "Do you agree?",
            True,
        ):
            return

        self._upload_on_exit = True
